@@ -3,14 +3,14 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2003 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2007 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
  * 
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  * 
  * GAWK is distributed in the hope that it will be useful,
@@ -20,7 +20,7 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
 /*
@@ -42,14 +42,14 @@ static int AVG_CHAIN_MAX = 2;	/* 11/2002: Modern machines are bigger, cut this d
 
 #include "awk.h"
 
-static NODE *assoc_find P((NODE *symbol, NODE *subs, int hash1));
+static NODE *assoc_find P((NODE *symbol, NODE *subs, unsigned long hash1));
 static void grow_table P((NODE *symbol));
 
-static unsigned long gst_hash_string P((const char *str, size_t len, unsigned long hsize));
+static unsigned long gst_hash_string P((const char *str, size_t len, unsigned long hsize, size_t *code));
 static unsigned long scramble P((unsigned long x));
-static unsigned long awk_hash P((const char *s, size_t len, unsigned long hsize));
+static unsigned long awk_hash P((const char *s, size_t len, unsigned long hsize, size_t *code));
 
-unsigned long (*hash)P((const char *s, size_t len, unsigned long hsize)) = awk_hash;
+unsigned long (*hash)P((const char *s, size_t len, unsigned long hsize, size_t *code)) = awk_hash;
 
 /* array_init --- possibly temporary function for experimentation purposes */
 
@@ -70,6 +70,192 @@ array_init()
 		hash = gst_hash_string; 
 }
 
+/*
+ * get_actual --- proceed to the actual Node_var_array,
+ *	change Node_var_new to an array.
+ *	If canfatal and type isn't good, die fatally,
+ *	otherwise return the final actual value.
+ */
+ 
+NODE *
+get_actual(NODE *symbol, int canfatal)
+{
+	int isparam = (symbol->type == Node_param_list
+			&& (symbol->flags & FUNC) == 0);
+	NODE *save_symbol = symbol;
+
+	if (isparam) {
+		save_symbol = symbol = stack_ptr[symbol->param_cnt];
+		if (symbol->type == Node_array_ref)
+			symbol = symbol->orig_array;
+	}
+
+	switch (symbol->type) {
+	case Node_var_new:
+		symbol->type = Node_var_array;
+		symbol->var_array = NULL;
+		/* fall through */
+	case Node_var_array:
+		break;
+
+	case Node_array_ref:
+	case Node_param_list:
+		if ((symbol->flags & FUNC) == 0)
+			cant_happen();
+		/* else
+			fall through */
+
+	default:
+		/* notably Node_var but catches also e.g. FS[1] = "x" */
+		if (canfatal) {
+			if ((symbol->flags & FUNC) != 0)
+				fatal(_("attempt to use function `%s' as an array"),
+								save_symbol->vname);
+			else if (isparam)
+				fatal(_("attempt to use scalar parameter `%s' as an array"),
+								save_symbol->vname);
+			else
+				fatal(_("attempt to use scalar `%s' as array"),
+								save_symbol->vname);
+		} else
+			break;
+	}
+
+	return symbol;
+}
+
+/*
+ * array_vname --- print the name of the array
+ *
+ * Returns a pointer to a statically maintained dynamically allocated string.
+ * It's appropriate for printing the name once; if the caller wants
+ * to save it, they have to make a copy.
+ *
+ * Setting MAX_LEN to a positive value (eg. 140) would limit the length
+ * of the output to _roughly_ that length.
+ *
+ * If MAX_LEN == 0, which is the default, the whole stack is printed.
+ */
+#define	MAX_LEN 0
+
+char *
+array_vname(register const NODE *symbol)
+{
+	if (symbol->type == Node_param_list)
+		symbol = stack_ptr[symbol->param_cnt];
+
+	if (symbol->type != Node_array_ref || symbol->orig_array->type != Node_var_array)
+		return symbol->vname;
+	else {
+		static char *message = NULL;
+		static size_t msglen = 0;
+		char *s;
+		size_t len;
+		int n;
+		const NODE *save_symbol = symbol;
+		const char *from = _("from %s");
+
+#if (MAX_LEN <= 0) || !defined(HAVE_SNPRINTF)
+		/* This is the default branch. */
+
+		/* First, we have to compute the length of the string: */
+		len = strlen(symbol->vname) + 2;	/* "%s (" */
+		n = 0;
+		do {
+			symbol = symbol->prev_array;
+			len += strlen(symbol->vname);
+			n++;
+		} while	(symbol->type == Node_array_ref);
+		/*
+		 * Each node contributes by strlen(from) minus the length
+		 * of "%s" in the translation (which is at least 2)
+		 * plus 2 for ", " or ")\0"; this adds up to strlen(from).
+		 */
+		len += n * strlen(from);
+
+		/* (Re)allocate memory: */
+		if (message == NULL) {
+			emalloc(message, char *, len, "array_vname");
+			msglen = len;
+		} else if (len > msglen) {
+			erealloc(message, char *, len, "array_vname");
+			msglen = len;
+		} /* else
+			current buffer can hold new name */
+
+		/* We're ready to print: */
+		symbol = save_symbol;
+		s = message;
+		/*
+		 * Ancient systems have sprintf() returning char *, not int.
+		 * Thus, `s += sprintf(s, from, name);' is a no-no.
+		 */
+		sprintf(s, "%s (", symbol->vname);
+		s += strlen(s);
+		for (;;) {
+			symbol = symbol->prev_array;
+			sprintf(s, from, symbol->vname);
+			s += strlen(s);
+			if (symbol->type != Node_array_ref)
+				break;
+			sprintf(s, ", ");
+			s += strlen(s);
+		}
+		sprintf(s, ")");
+
+#else /* MAX_LEN > 0 */
+
+		/*
+		 * The following check fails only on
+		 * abnormally_long_variable_name.
+		 */
+#define PRINT_CHECK \
+		if (n <= 0 || n >= len) \
+			return save_symbol->vname; \
+		s += n; len -= n
+#define PRINT(str) \
+		n = snprintf(s, len, str); \
+		PRINT_CHECK
+#define PRINT_vname(str) \
+		n = snprintf(s, len, str, symbol->vname); \
+		PRINT_CHECK
+
+		if (message == NULL)
+			emalloc(message, char *, MAX_LEN, "array_vname");
+
+		s = message;
+		len = MAX_LEN;
+
+		/* First, print the vname of the node. */
+		PRINT_vname("%s (");
+
+		for (;;) {
+			symbol = symbol->prev_array;
+			/*
+			 * When we don't have enough space and this is not
+			 * the last node, shorten the list.
+			 */
+			if (len < 40 && symbol->type == Node_array_ref) {
+				PRINT("..., ");
+				symbol = symbol->orig_array;
+			}
+			PRINT_vname(from);
+			if (symbol->type != Node_array_ref)
+				break;
+			PRINT(", ");
+		}
+		PRINT(")");
+
+#undef PRINT_CHECK
+#undef PRINT
+#undef PRINT_vname
+#endif /* MAX_LEN <= 0 */
+
+		return message;
+	}
+}
+#undef MAX_LEN
+
 /* concat_exp --- concatenate expression list into a single string */
 
 NODE *
@@ -121,7 +307,7 @@ concat_exp(register NODE *tree)
 void
 assoc_clear(NODE *symbol)
 {
-	int i;
+	long i;
 	NODE *bucket, *next;
 
 	if (symbol->var_array == NULL)
@@ -140,10 +326,10 @@ assoc_clear(NODE *symbol)
 	symbol->flags &= ~ARRAYMAXED;
 }
 
-/* hash --- calculate the hash function of the string in subs */
+/* awk_hash --- calculate the hash function of the string in subs */
 
 static unsigned long
-awk_hash(register const char *s, register size_t len, unsigned long hsize)
+awk_hash(register const char *s, register size_t len, unsigned long hsize, size_t *code)
 {
 	register unsigned long h = 0;
 
@@ -221,6 +407,8 @@ awk_hash(register const char *s, register size_t len, unsigned long hsize)
 		}
 	}
 #endif /* ! VAXC */
+	if (code != NULL)
+		*code = h;
 
 	if (h >= hsize)
 		h %= hsize;
@@ -230,7 +418,7 @@ awk_hash(register const char *s, register size_t len, unsigned long hsize)
 /* assoc_find --- locate symbol[subs] */
 
 static NODE *				/* NULL if not found */
-assoc_find(NODE *symbol, register NODE *subs, int hash1)
+assoc_find(NODE *symbol, register NODE *subs, unsigned long hash1)
 {
 	register NODE *bucket;
 	const char *s1_str;
@@ -249,7 +437,7 @@ assoc_find(NODE *symbol, register NODE *subs, int hash1)
 
 		if (s1_len == s2->stlen) {
 			if (s1_len == 0		/* "" is a valid index */
-			    || STREQN(s1_str, s2->stptr, s1_len))
+			    || memcmp(s1_str, s2->stptr, s1_len) == 0)
 				return bucket;
 		}
 	}
@@ -263,24 +451,20 @@ assoc_find(NODE *symbol, register NODE *subs, int hash1)
 NODE *
 in_array(NODE *symbol, NODE *subs)
 {
-	register int hash1;
+	register unsigned long hash1;
 	NODE *ret;
 
-	if (symbol->type == Node_param_list)
-		symbol = stack_ptr[symbol->param_cnt];
-	if (symbol->type == Node_array_ref)
-		symbol = symbol->orig_array;
-	if ((symbol->flags & SCALAR) != 0)
-		fatal(_("attempt to use scalar `%s' as array"), symbol->vname);
+	symbol = get_array(symbol);
+
 	/*
-	 * evaluate subscript first, it could have side effects
+	 * Evaluate subscript first, it could have side effects.
 	 */
 	subs = concat_exp(subs);	/* concat_exp returns a string node */
 	if (symbol->var_array == NULL) {
 		free_temp(subs);
 		return NULL;
 	}
-	hash1 = hash(subs->stptr, subs->stlen, (unsigned long) symbol->array_size);
+	hash1 = hash(subs->stptr, subs->stlen, (unsigned long) symbol->array_size, NULL);
 	ret = assoc_find(symbol, subs, hash1);
 	free_temp(subs);
 	if (ret)
@@ -301,31 +485,23 @@ in_array(NODE *symbol, NODE *subs)
 NODE **
 assoc_lookup(NODE *symbol, NODE *subs, int reference)
 {
-	register int hash1;
+	register unsigned long hash1;
 	register NODE *bucket;
+	size_t code;
 
-	/* protect against silly users, e.g. FS[1] = "x" */
-	if (symbol->type != Node_var_array && symbol->type != Node_var)
-		fatal(_("attempt to use scalar `%s' as array"), symbol->vname);
+	assert(symbol->type == Node_var_array);
 
 	(void) force_string(subs);
 
-	if ((symbol->flags & SCALAR) != 0)
-		fatal(_("attempt to use scalar `%s' as array"), symbol->vname);
-
 	if (symbol->var_array == NULL) {
-		if (symbol->type != Node_var_array) {
-			unref(symbol->var_value);
-			symbol->type = Node_var_array;
-		}
 		symbol->array_size = symbol->table_size = 0;	/* sanity */
 		symbol->flags &= ~ARRAYMAXED;
 		grow_table(symbol);
 		hash1 = hash(subs->stptr, subs->stlen,
-				(unsigned long) symbol->array_size);
+				(unsigned long) symbol->array_size, & code);
 	} else {
 		hash1 = hash(subs->stptr, subs->stlen,
-				(unsigned long) symbol->array_size);
+				(unsigned long) symbol->array_size, & code);
 		bucket = assoc_find(symbol, subs, hash1);
 		if (bucket != NULL) {
 			free_temp(subs);
@@ -336,13 +512,13 @@ assoc_lookup(NODE *symbol, NODE *subs, int reference)
 	if (do_lint && reference) {
 		subs->stptr[subs->stlen] = '\0';
 		lintwarn(_("reference to uninitialized element `%s[\"%s\"]'"),
-		      symbol->vname, subs->stptr);
+		      array_vname(symbol), subs->stptr);
 	}
 
 	/* It's not there, install it. */
 	if (do_lint && subs->stlen == 0)
 		lintwarn(_("subscript of array `%s' is null string"),
-			symbol->vname);
+			array_vname(symbol));
 
 	/* first see if we would need to grow the array, before installing */
 	symbol->table_size++;
@@ -350,8 +526,7 @@ assoc_lookup(NODE *symbol, NODE *subs, int reference)
 	    && (symbol->table_size / symbol->array_size) > AVG_CHAIN_MAX) {
 		grow_table(symbol);
 		/* have to recompute hash value for new size */
-		hash1 = hash(subs->stptr, subs->stlen,
-				(unsigned long) symbol->array_size);
+		hash1 = code % (unsigned long) symbol->array_size;
 	}
 
 	getnode(bucket);
@@ -367,16 +542,24 @@ assoc_lookup(NODE *symbol, NODE *subs, int reference)
 	 */
 	bucket->flags |= MALLOC;
 	bucket->ahname_ref = 1;
-	emalloc(bucket->ahname_str, char *, subs->stlen + 2, "assoc_lookup");
-	bucket->ahname_len = subs->stlen;
 
-	memcpy(bucket->ahname_str, subs->stptr, subs->stlen);
-	bucket->ahname_str[bucket->ahname_len] = '\0';
-
-	free_temp(subs);
+	/* For TEMP node, reuse the storage directly */
+	if ((subs->flags & TEMP) != 0) {
+		bucket->ahname_str = subs->stptr;
+		bucket->ahname_len = subs->stlen;
+		bucket->ahname_str[bucket->ahname_len] = '\0';
+		subs->flags &= ~TEMP;   /* for good measure */
+		freenode(subs);
+	} else {
+		emalloc(bucket->ahname_str, char *, subs->stlen + 2, "assoc_lookup");
+		bucket->ahname_len = subs->stlen;
+		memcpy(bucket->ahname_str, subs->stptr, subs->stlen);
+		bucket->ahname_str[bucket->ahname_len] = '\0';
+	}
 
 	bucket->ahvalue = Nnull_string;
 	bucket->ahnext = symbol->var_array[hash1];
+	bucket->ahcode = code;
 	symbol->var_array[hash1] = bucket;
 	return &(bucket->ahvalue);
 }
@@ -389,86 +572,63 @@ assoc_lookup(NODE *symbol, NODE *subs, int reference)
  */
 
 void
-do_delete(NODE *symbol, NODE *tree)
+do_delete(NODE *sym, NODE *tree)
 {
-	register int hash1;
+	register unsigned long hash1;
 	register NODE *bucket, *last;
 	NODE *subs;
-
-	/*
-	 * Evaluate subscript first, always, in case there are
-	 * side effects.
-	 */
-	if (tree != NULL)
-		subs = concat_exp(tree);	/* concat_exp returns string node */
-	else
-		subs = NULL;
-
-	if (symbol->type == Node_param_list) {
-		symbol = stack_ptr[symbol->param_cnt];
-		if (symbol->type == Node_var) {
-			if (subs != NULL) {
-				if (do_lint)
-					lintwarn(_("delete: index `%s' not in array `%s'"),
-						subs->stptr, symbol->vname);
-				free_temp(subs);
-			}
-			return;
-		}
-	}
-	if (symbol->type == Node_array_ref)
-		symbol = symbol->orig_array;
-	if (symbol->type == Node_var_array) {
-		if (symbol->var_array == NULL) {
-			if (subs != NULL) {
-				if (do_lint)
-					lintwarn(_("delete: index `%s' not in array `%s'"),
-						subs->stptr, symbol->vname);
-				free_temp(subs);
-			}
-			return;
-		}
-	} else
-		fatal(_("delete: illegal use of variable `%s' as array"),
-			symbol->vname);
+	register NODE *symbol = get_array(sym);
 
 	if (tree == NULL) {	/* delete array */
 		assoc_clear(symbol);
 		return;
 	}
 
-	hash1 = hash(subs->stptr, subs->stlen, (unsigned long) symbol->array_size);
+	last = NULL;	/* shut up gcc -Wall */
+	hash1 = 0;	/* ditto */
 
-	last = NULL;
-	for (bucket = symbol->var_array[hash1]; bucket != NULL;
-			last = bucket, bucket = bucket->ahnext) {
-		/*
-		 * This used to use cmp_nodes() here.  That's wrong.
-		 * Array indexes are strings; compare as such, always!
-		 */
-		const char *s1_str;
-		size_t s1_len;
-		NODE *s2;
+	/*
+	 * Always evaluate subscript, it could have side effects.
+	 */
+	subs = concat_exp(tree);	/* concat_exp returns string node */
 
-		s1_str = bucket->ahname_str;
-		s1_len = bucket->ahname_len;
-		s2 = subs;
+	if (symbol->var_array != NULL) {
+		hash1 = hash(subs->stptr, subs->stlen,
+				(unsigned long) symbol->array_size, NULL);
+		last = NULL;
+		for (bucket = symbol->var_array[hash1]; bucket != NULL;
+				last = bucket, bucket = bucket->ahnext) {
+			/*
+			 * This used to use cmp_nodes() here.  That's wrong.
+			 * Array indexes are strings; compare as such, always!
+			 */
+			const char *s1_str;
+			size_t s1_len;
+			NODE *s2;
 
-		if (s1_len == s2->stlen) {
-			if (s1_len == 0		/* "" is a valid index */
-			    || STREQN(s1_str, s2->stptr, s1_len))
-				break;
+			s1_str = bucket->ahname_str;
+			s1_len = bucket->ahname_len;
+			s2 = subs;
+	
+			if (s1_len == s2->stlen) {
+				if (s1_len == 0		/* "" is a valid index */
+				    || memcmp(s1_str, s2->stptr, s1_len) == 0)
+					break;
+			}
 		}
-	}
+	} else
+		bucket = NULL;	/* The array is empty.  */
 
 	if (bucket == NULL) {
 		if (do_lint)
 			lintwarn(_("delete: index `%s' not in array `%s'"),
-				subs->stptr, symbol->vname);
+				subs->stptr, array_vname(sym));
 		free_temp(subs);
 		return;
 	}
+
 	free_temp(subs);
+
 	if (last != NULL)
 		last->ahnext = bucket->ahnext;
 	else
@@ -497,23 +657,14 @@ do_delete(NODE *symbol, NODE *tree)
 void
 do_delete_loop(NODE *symbol, NODE *tree)
 {
-	size_t i;
+	long i;
 	NODE **lhs;
 	Func_ptr after_assign = NULL;
 
-	if (symbol->type == Node_param_list) {
-		symbol = stack_ptr[symbol->param_cnt];
-		if (symbol->type == Node_var)
-			return;
-	}
-	if (symbol->type == Node_array_ref)
-		symbol = symbol->orig_array;
-	if (symbol->type == Node_var_array) {
-		if (symbol->var_array == NULL)
-			return;
-	} else
-		fatal(_("delete: illegal use of variable `%s' as array"),
-			symbol->vname);
+	symbol = get_array(symbol);
+
+	if (symbol->var_array == NULL)
+		return;
 
 	/* get first index value */
 	for (i = 0; i < symbol->array_size; i++) {
@@ -540,7 +691,7 @@ grow_table(NODE *symbol)
 	NODE **old, **new, *chain, *next;
 	int i, j;
 	unsigned long hash1;
-	unsigned long oldsize, newsize;
+	unsigned long oldsize, newsize, k;
 	/*
 	 * This is an array of primes. We grow the table by an order of
 	 * magnitude each time (not just doubling) so that growing is a
@@ -585,14 +736,13 @@ grow_table(NODE *symbol)
 
 	/* old hash table there, move stuff to new, free old */
 	old = symbol->var_array;
-	for (i = 0; i < oldsize; i++) {
-		if (old[i] == NULL)
+	for (k = 0; k < oldsize; k++) {
+		if (old[k] == NULL)
 			continue;
 
-		for (chain = old[i]; chain != NULL; chain = next) {
+		for (chain = old[k]; chain != NULL; chain = next) {
 			next = chain->ahnext;
-			hash1 = hash(chain->ahname_str,
-					chain->ahname_len, newsize);
+			hash1 = chain->ahcode % newsize;
 
 			/* remove from old list, add to new */
 			chain->ahnext = new[hash1];
@@ -610,6 +760,16 @@ done:
 	symbol->array_size = newsize;
 }
 
+/* set_SUBSEP --- make sure SUBSEP always has a string value */
+
+void
+set_SUBSEP(void)
+{
+
+	(void) force_string(SUBSEP_node->var_value);
+	return;
+}
+
 /* pr_node --- print simple node info */
 
 static void
@@ -626,7 +786,7 @@ pr_node(NODE *n)
 NODE *
 assoc_dump(NODE *symbol)
 {
-	int i;
+	long i;
 	NODE *bucket;
 
 	if (symbol->var_array == NULL) {
@@ -695,7 +855,7 @@ static void
 dup_table(NODE *symbol, NODE *newsymb)
 {
 	NODE **old, **new, *chain, *bucket;
-	int i;
+	long i;
 	unsigned long cursize;
 
 	/* find the current hash size */
@@ -793,7 +953,7 @@ merge(NODE *left, NODE *right)
 /* merge_sort --- recursively sort the left and right sides of a list */
 
 static NODE *
-merge_sort(NODE *left, int size)
+merge_sort(NODE *left, unsigned long size)
 {
 	NODE *right, *tmp;
 	int i, half;
@@ -830,10 +990,12 @@ assoc_from_list(NODE *symbol, NODE *list)
 {
 	NODE *next;
 	unsigned long i = 0;
-	register int hash1;
+	register unsigned long hash1;
 	char buf[100];
 
 	for (; list != NULL; list = next) {
+		size_t code;
+
 		next = list->ahnext;
 
 		/* make an int out of i++ */
@@ -847,7 +1009,8 @@ assoc_from_list(NODE *symbol, NODE *list)
 
 		/* find the bucket where it belongs */
 		hash1 = hash(list->ahname_str, list->ahname_len,
-				symbol->array_size);
+				symbol->array_size, & code);
+		list->ahcode = code;
 
 		/* link the node into the chain at that bucket */
 		list->ahnext = symbol->var_array[hash1];
@@ -865,7 +1028,7 @@ typedef enum asort_how { VALUE, INDEX } ASORT_TYPE;
 static NODE *
 assoc_sort_inplace(NODE *symbol, ASORT_TYPE how)
 {
-	int i, num;
+	unsigned long i, num;
 	NODE *bucket, *next, *list;
 
 	if (symbol->var_array == NULL
@@ -962,32 +1125,17 @@ assoc_sort_inplace(NODE *symbol, ASORT_TYPE how)
 static NODE *
 asort_actual(NODE *tree, ASORT_TYPE how)
 {
-	NODE *src, *dest;
-
-	src = tree->lnode;
-	dest = NULL;
-
-	if (src->type == Node_param_list)
-		src = stack_ptr[src->param_cnt];
-	if (src->type == Node_array_ref)
-		src = src->orig_array;
-	if (src->type != Node_var_array)
-		fatal(_("asort: first argument is not an array"));
+	NODE *array = get_array(tree->lnode);
 
 	if (tree->rnode != NULL) {  /* 2nd optional arg */
-		dest = tree->rnode->lnode;
-		if (dest->type == Node_param_list)
-			dest = stack_ptr[dest->param_cnt];
-		if (dest->type == Node_array_ref)
-			dest = dest->orig_array;
-		if (dest->type != Node_var && dest->type != Node_var_array)
-			fatal(_("asort: second argument is not an array"));
-		dest->type = Node_var_array;
+		NODE *dest = get_array(tree->rnode->lnode);
+
 		assoc_clear(dest);
-		dup_table(src, dest);
+		dup_table(array, dest);
+		array = dest;
 	}
 
-	return dest != NULL ? assoc_sort_inplace(dest, how) : assoc_sort_inplace(src, how);
+	return assoc_sort_inplace(array, how);
 }
 
 /* do_asort --- sort array by value */
@@ -1030,7 +1178,7 @@ Paolo
  */
 
 static unsigned long
-gst_hash_string(const char *str, size_t len, unsigned long hsize)
+gst_hash_string(const char *str, size_t len, unsigned long hsize, size_t *code)
 {
 	unsigned long hashVal = 1497032417;    /* arbitrary value */
 	unsigned long ret;
@@ -1042,6 +1190,10 @@ gst_hash_string(const char *str, size_t len, unsigned long hsize)
 	}
 
 	ret = scramble(hashVal);
+
+	if (code != NULL)
+		*code = ret;
+
 	if (ret >= hsize)
 		ret %= hsize;
 
