@@ -3,14 +3,14 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2004 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2007 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
  * 
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  * 
  * GAWK is distributed in the hope that it will be useful,
@@ -20,9 +20,12 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
+/* FIX THIS BEFORE EVERY RELEASE: */
+#define UPDATE_YEAR	2007
+
 #include "awk.h"
 #include "getopt.h"
 
@@ -44,6 +47,8 @@ static void copyleft P((void)) ATTRIBUTE_NORETURN;
 static void cmdline_fs P((char *str));
 static void init_args P((int argc0, int argc, char *argv0, char **argv));
 static void init_vars P((void));
+static NODE *load_environ P((void));
+static NODE *load_procinfo P((void));
 static void add_src P((struct src **data, long *num, long *alloc, enum srctype stype, char *val));
 static RETSIGTYPE catchsig P((int sig)) ATTRIBUTE_NORETURN;
 static void nostalgia P((void)) ATTRIBUTE_NORETURN;
@@ -82,9 +87,9 @@ int errcount = 0;		/* error counter, used by yyerror() */
 
 NODE *Nnull_string;		/* The global null string */
 
-#if ENABLE_NLS && defined(HAVE_LOCALE_H)
+#if defined(HAVE_LOCALE_H)
 struct lconv loc;		/* current locale */
-#endif /* ENABLE_NLS && defined(HAVE_LOCALE_H) */
+#endif /* defined(HAVE_LOCALE_H) */
 
 /* The name the program was invoked under, for error messages */
 const char *myname;
@@ -113,6 +118,8 @@ static struct src *preassigns = NULL;	/* requested via -v or -F */
 static long numassigns = -1;		/* how many of them */
 static long allocassigns;		/* for how many is allocated */
 
+static int disallow_var_assigns = FALSE;	/* true for --exec */
+
 #define	preassigns_add(stype, val) \
 	add_src(&preassigns, &numassigns, &allocassigns, stype, val)
 
@@ -132,10 +139,13 @@ int do_dump_vars = FALSE;	/* dump all global variables at end */
 int do_tidy_mem = FALSE;	/* release vars when done */
 
 int in_begin_rule = FALSE;	/* we're in a BEGIN rule */
-int in_end_rule = FALSE;	/* we're in a END rule */
+int in_end_rule = FALSE;	/* we're in an END rule */
 int whiny_users = FALSE;	/* do things that whiny users want */
+int use_lc_numeric = FALSE;	/* obey locale for decimal point */
 #ifdef MBS_SUPPORT
-int gawk_mb_cur_max = 1;	/* MB_CUR_MAX value, see comment in main() */
+int gawk_mb_cur_max;		/* MB_CUR_MAX value, see comment in main() */
+#else
+const int gawk_mb_cur_max = 1;
 #endif
 
 int output_is_tty = FALSE;	/* control flushing of output */
@@ -153,7 +163,7 @@ NODE *expression_value;
 #if _MSC_VER == 510
 void (*lintfunc) P((va_list va_alist, ...)) = warning;
 #else
-#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
+#ifdef CAN_USE_STDARG_H
 void (*lintfunc) P((const char *mesg, ...)) = warning;
 #else
 void (*lintfunc) () = warning;
@@ -181,7 +191,9 @@ static const struct option optab[] = {
 	{ "version",		no_argument,		NULL,		'V' },
 	{ "usage",		no_argument,		NULL,		'u' },
 	{ "help",		no_argument,		NULL,		'u' },
-#ifdef GAWKDEBUG
+	{ "exec",		required_argument,	NULL,		'S' },
+	{ "use-lc-numeric",	no_argument,		& use_lc_numeric, 1 },
+#if defined(YYDEBUG) || defined(GAWKDEBUG)
 	{ "parsedebug",		no_argument,		NULL,		'D' },
 #endif
 	{ NULL, 0, NULL, '\0' }
@@ -207,6 +219,7 @@ main(int argc, char **argv)
 	extern int opterr;
 	extern char *optarg;
 	int i;
+	int stdio_problem = FALSE;
 
 	/* do these checks early */
 	if (getenv("TIDYMEM") != NULL)
@@ -216,17 +229,19 @@ main(int argc, char **argv)
 		whiny_users = TRUE;
 
 #ifdef HAVE_MCHECK_H
+#ifdef HAVE_MTRACE
 	if (do_tidy_mem)
 		mtrace();
+#endif /* HAVE_MTRACE */
 #endif /* HAVE_MCHECK_H */
-	
+
 #if defined(LC_CTYPE)
 	setlocale(LC_CTYPE, "");
 #endif
 #if defined(LC_COLLATE)
 	setlocale(LC_COLLATE, "");
 #endif
-#if HAVE_LC_MESSAGES && defined(LC_MESSAGES)
+#if defined(LC_MESSAGES)
 	setlocale(LC_MESSAGES, "");
 #endif
 #if defined(LC_NUMERIC)
@@ -235,7 +250,16 @@ main(int argc, char **argv)
 	 * point is used for parsing source code and for command-line
 	 * assignments and the locale value for processing input,
 	 * number to string conversion, and printing output.
+	 *
+	 * 10/2005 --- see below also; we now only use the locale's
+	 * decimal point if do_posix in effect.
+	 *
+	 * 9/2007:
+	 * This is a mess. We need to get the locale's numeric info for
+	 * the thousands separator for the %'d flag.
 	 */
+	setlocale(LC_NUMERIC, "");
+	loc = *localeconv();	/* Make a local copy of locale numeric info, early on */
 	setlocale(LC_NUMERIC, "C");
 #endif
 #if defined(LC_TIME)
@@ -249,6 +273,7 @@ main(int argc, char **argv)
 	 * this value once makes a speed difference.
 	 */
 	gawk_mb_cur_max = MB_CUR_MAX;
+	/* Without MBS_SUPPORT, gawk_mb_cur_max is 1. */
 #endif
 
 	(void) bindtextdomain(PACKAGE, LOCALEDIR);
@@ -292,6 +317,9 @@ main(int argc, char **argv)
 			preassigns_add(PRE_ASSIGN_FS, optarg);
 			break;
 
+		case 'S':
+			disallow_var_assigns = TRUE;
+			/* fall through */
 		case 'f':
 			/*
 			 * a la MKS awk, allow multiple -f options.
@@ -398,7 +426,7 @@ main(int argc, char **argv)
 			break;
 
 		case 'D':
-#ifdef GAWKDEBUG
+#if defined(YYDEBUG) || defined(GAWKDEBUG)
 			yydebug = 2;
 			break;
 #endif
@@ -436,6 +464,8 @@ main(int argc, char **argv)
 				let getopt print error message for us */
 			break;
 		}
+		if (c == 'S')	/* --exec ends option processing */
+			break;
 	}
 out:
 
@@ -451,6 +481,7 @@ out:
 	}
 
 	if (do_posix) {
+		use_lc_numeric = TRUE;
 		if (do_traditional)	/* both on command line */
 			warning(_("`--posix' overrides `--traditional'"));
 		else
@@ -532,6 +563,15 @@ out:
 	init_args(optind, argc, (char *) myname, argv);
 	(void) tokexpand();
 
+#if defined(LC_NUMERIC)
+	/*
+	 * FRAGILE!  CAREFUL!
+	 * Pre-initing the variables with arg_assign() can change the
+	 * locale.  Force it to C before parsing the program.
+	 */
+	setlocale(LC_NUMERIC, "C");
+#endif
+
 	/* Read in the program */
 	if (yyparse() != 0 || errcount != 0)
 		exit(1);
@@ -551,12 +591,23 @@ out:
 	init_profiling_signals();
 
 #if defined(LC_NUMERIC)
-	/* See comment above. */
-	setlocale(LC_NUMERIC, "");
-#endif
-
-#if ENABLE_NLS && defined(HAVE_LOCALE_H)
-	loc = *localeconv();	/* Make a local copy of locale numeric info */
+	/*
+	 * See comment above about using locale's decimal point.
+	 *
+	 * 10/2005:
+	 * Bitter experience teaches us that most people the world over
+	 * use period as the decimal point, not whatever their locale
+	 * uses.  Thus, only use the locale's decimal point if being
+	 * posixly anal-retentive.
+	 *
+	 * 7/2007:
+	 * Be a little bit kinder. Allow the --use-lc-numeric option
+	 * to also use the local decimal point. This avoids the draconian
+	 * strictness of POSIX mode if someone just wants to parse their
+	 * data using the local decimal point.
+	 */
+	if (use_lc_numeric)
+		setlocale(LC_NUMERIC, "");
 #endif
 
 	/* Whew. Finally, run the program. */
@@ -572,7 +623,23 @@ out:
 		(void) interpret(end_block);
 	}
 	in_end_rule = FALSE;
-	if (close_io() != 0 && ! exiting && exit_val == 0)
+	/*
+	 * This used to be:
+	 *
+	 * if (close_io() != 0 && ! exiting && exit_val == 0)
+	 * 	exit_val = 1;
+	 *
+	 * Other awks don't care about problems closing open files
+	 * and pipes, in that it doesn't affect their exit status.
+	 * So we no longer do either.
+	 */
+	(void) close_io(& stdio_problem);
+	/*
+	 * However, we do want to exit non-zero if there was a problem
+	 * with stdout/stderr, so we reinstate a slightly different
+	 * version of the above:
+	 */
+	if (stdio_problem && ! exiting && exit_val == 0)
 		exit_val = 1;
 
 	if (do_profiling) {
@@ -636,6 +703,7 @@ usage(int exitval, FILE *fp)
 	fputs(_("\t-W copyleft\t\t--copyleft\n"), fp);
 	fputs(_("\t-W copyright\t\t--copyright\n"), fp);
 	fputs(_("\t-W dump-variables[=file]\t--dump-variables[=file]\n"), fp);
+	fputs(_("\t-W exec=file\t\t--exec=file\n"), fp);
 	fputs(_("\t-W gen-po\t\t--gen-po\n"), fp);
 	fputs(_("\t-W help\t\t\t--help\n"), fp);
 	fputs(_("\t-W lint[=fatal]\t\t--lint[=fatal]\n"), fp);
@@ -653,10 +721,16 @@ usage(int exitval, FILE *fp)
 	fputs(_("\t-W source=program-text\t--source=program-text\n"), fp);
 	fputs(_("\t-W traditional\t\t--traditional\n"), fp);
 	fputs(_("\t-W usage\t\t--usage\n"), fp);
+	fputs(_("\t-W use-lc-numeric\t--use-lc-numeric\n"), fp);
 	fputs(_("\t-W version\t\t--version\n"), fp);
 
 
 	/* This is one string to make things easier on translators. */
+	/* TRANSLATORS: --help output 5 (end)
+	   TRANSLATORS: the placeholder indicates the bug-reporting address
+	   for this application.  Please add _another line_ with the
+	   address for translation bugs.
+	   no-wrap */
 	fputs(_("\nTo report bugs, see node `Bugs' in `gawk.info', which is\n\
 section `Reporting Problems and Bugs' in the printed version.\n\n"), fp);
 
@@ -670,8 +744,11 @@ By default it reads standard input and writes standard output.\n\n"), fp);
 
 	fflush(fp);
 
-	if (ferror(fp))
+	if (ferror(fp)) {
+		if (fp == stdout)
+			warning(_("error writing standard output (%s)"), strerror(errno));
 		exit(1);
+	}
 
 	exit(exitval);
 }
@@ -686,7 +763,7 @@ copyleft()
 \n\
 This program is free software; you can redistribute it and/or modify\n\
 it under the terms of the GNU General Public License as published by\n\
-the Free Software Foundation; either version 2 of the License, or\n\
+the Free Software Foundation; either version 3 of the License, or\n\
 (at your option) any later version.\n\
 \n");
 	static const char blurb_part2[] =
@@ -696,18 +773,24 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
 GNU General Public License for more details.\n\
 \n");
 	static const char blurb_part3[] =
+/*
 	  N_("You should have received a copy of the GNU General Public License\n\
 along with this program; if not, write to the Free Software\n\
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.\n");
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.\n");
+*/
+	  N_("You should have received a copy of the GNU General Public License\n\
+along with this program. If not, see http://www.gnu.org/licenses/.\n");
  
 	/* multiple blurbs are needed for some brain dead compilers. */
-	printf(_(blurb_part1), 2003);	/* Last update year */
+	printf(_(blurb_part1), UPDATE_YEAR);	/* Last update year */
 	fputs(_(blurb_part2), stdout);
 	fputs(_(blurb_part3), stdout);
 	fflush(stdout);
 
-	if (ferror(stdout))
+	if (ferror(stdout)) {
+		warning(_("error writing standard output (%s)"), strerror(errno));
 		exit(1);
+	}
 
 	exit(0);
 }
@@ -816,11 +899,16 @@ init_vars()
 		if (vp->assign)
 			(*(vp->assign))();
 	}
+
+	/* Set up deferred variables (loaded only when accessed). */
+	if (! do_traditional)
+		register_deferred_variable("PROCINFO", load_procinfo);
+	register_deferred_variable("ENVIRON", load_environ);
 }
 
 /* load_environ --- populate the ENVIRON array */
 
-NODE *
+static NODE *
 load_environ()
 {
 #if ! defined(TANDEM)
@@ -857,9 +945,16 @@ load_environ()
 	 * Put AWKPATH into ENVIRON if it's not there.
 	 * This allows querying it from within awk programs.
 	 */
-	if (getenv("AWKPATH") == NULL) {
+	if (! in_array(ENVIRON_node, tmp_string("AWKPATH", 7))) {
+		/*
+		 * On VMS, environ[] only holds a subset of what getenv() can
+		 * find, so look AWKPATH up before resorting to default path.
+		 */
+		val = getenv("AWKPATH");
+		if (val == NULL)
+			val = defpath;
 		aptr = assoc_lookup(ENVIRON_node, tmp_string("AWKPATH", 7), FALSE);
-		*aptr = make_string(defpath, strlen(defpath));
+		*aptr = make_string(val, strlen(val));
 	}
 #endif /* TANDEM */
 	return ENVIRON_node;
@@ -867,7 +962,7 @@ load_environ()
 
 /* load_procinfo --- populate the PROCINFO array */
 
-NODE *
+static NODE *
 load_procinfo()
 {
 	int i;
@@ -952,6 +1047,9 @@ arg_assign(char *arg, int initing)
 	NODE *it;
 	NODE **lhs;
 
+	if (! initing && disallow_var_assigns)
+		return FALSE;	/* --exec */
+
 	cp = strchr(arg, '=');
 
 	if (cp == NULL) {
@@ -992,9 +1090,14 @@ arg_assign(char *arg, int initing)
 		it = make_str_node(cp, strlen(cp), SCAN);
 		it->flags |= MAYBE_NUM;
 #ifdef LC_NUMERIC
-		setlocale(LC_NUMERIC, "C");
+		/*
+		 * See comment above about locale decimal point.
+		 */
+		if (do_posix)
+			setlocale(LC_NUMERIC, "C");
 		(void) force_number(it);
-		setlocale(LC_NUMERIC, "");
+		if (do_posix)
+			setlocale(LC_NUMERIC, "");
 #endif /* LC_NUMERIC */
 		var = variable(arg, FALSE, Node_var);
 		lhs = get_lhs(var, &after_assign, FALSE);
@@ -1004,7 +1107,8 @@ arg_assign(char *arg, int initing)
 			(*after_assign)();
 	}
 
-	*--cp = '=';	/* restore original text of ARGV */
+	if (! initing)
+		*--cp = '=';	/* restore original text of ARGV */
 
 	return ! badvar;
 }
@@ -1066,6 +1170,7 @@ init_fds()
 	struct stat sbuf;
 	int fd;
 	int newfd;
+	char const *const opposite_mode[] = {"w", "r", "r"};
 
 	/* maybe no stderr, don't bother with error mesg */
 	for (fd = 0; fd <= 2; fd++) {
@@ -1074,7 +1179,7 @@ init_fds()
 			if (do_lint)
 				lintwarn(_("no pre-opened fd %d"), fd);
 #endif
-			newfd = devopen("/dev/null", "r+");
+			newfd = devopen("/dev/null", opposite_mode[fd]);
 			/* turn off some compiler warnings "set but not used" */
 			newfd += 0;
 #ifdef MAKE_A_HEROIC_EFFORT
