@@ -3,7 +3,7 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2001 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2001, 2003 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -20,7 +20,7 @@
  * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
 #include "awk.h"
@@ -28,12 +28,9 @@
 int sourceline = 0;
 char *source = NULL;
 
-static char *srcfile = NULL;
+static const char *srcfile = NULL;
 static int srcline;
 
-/* prototype needed for ansi / gcc */
-void err P((const char *s, const char *emsg, va_list argp));
-
 /* err --- print an error message with source line and file and record */
 
 /* VARARGS2 */
@@ -65,7 +62,7 @@ err(const char *s, const char *emsg, va_list argp)
 			(void) fprintf(stderr, "FILENAME=%s ", file);
 		(void) fprintf(stderr, "FNR=%ld) ", FNR);
 	}
-	(void) fprintf(stderr, s);
+	(void) fprintf(stderr, "%s", s);
 	vfprintf(stderr, emsg, argp);
 	(void) fprintf(stderr, "\n");
 	(void) fflush(stderr);
@@ -79,8 +76,8 @@ err(const char *s, const char *emsg, va_list argp)
  */
 
 void
-#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
-  msg(char *mesg, ...)
+#ifdef CAN_USE_STDARG_H
+  msg(const char *mesg, ...)
 #else
 /*VARARGS0*/
   msg(va_alist)
@@ -88,7 +85,7 @@ void
 #endif
 {
 	va_list args;
-#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
+#ifdef CAN_USE_STDARG_H
 	va_start(args, mesg);
 #else
 	char *mesg;
@@ -103,8 +100,8 @@ void
 /* warning --- print a warning message */
 
 void
-#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
-  warning(char *mesg, ...)
+#ifdef CAN_USE_STDARG_H
+  warning(const char *mesg, ...)
 #else
 /*VARARGS0*/
   warning(va_alist)
@@ -112,7 +109,7 @@ void
 #endif
 {
 	va_list args;
-#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
+#ifdef CAN_USE_STDARG_H
 	va_start(args, mesg);
 #else
 	char *mesg;
@@ -125,8 +122,8 @@ void
 }
 
 void
-#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
-  error(char *mesg, ...)
+#ifdef CAN_USE_STDARG_H
+  error(const char *mesg, ...)
 #else
 /*VARARGS0*/
   error(va_alist)
@@ -134,7 +131,7 @@ void
 #endif
 {
 	va_list args;
-#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
+#ifdef CAN_USE_STDARG_H
 	va_start(args, mesg);
 #else
 	char *mesg;
@@ -149,17 +146,20 @@ void
 /* set_loc --- set location where a fatal error happened */
 
 void
-set_loc(char *file, int line)
+set_loc(const char *file, int line)
 {
 	srcfile = file;
 	srcline = line;
+
+	/* This stupid line keeps some compilers happy: */
+	file = srcfile; line = srcline;
 }
 
 /* fatal --- print an error message and die */
 
 void
-#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
-  r_fatal(char *mesg, ...)
+#ifdef CAN_USE_STDARG_H
+  r_fatal(const char *mesg, ...)
 #else
 /*VARARGS0*/
   r_fatal(va_alist)
@@ -167,7 +167,7 @@ void
 #endif
 {
 	va_list args;
-#if defined(HAVE_STDARG_H) && defined(__STDC__) && __STDC__
+#ifdef CAN_USE_STDARG_H
 	va_start(args, mesg);
 #else
 	char *mesg;
